/* this is the O(sqrt(N)) storage merge algorithm extracted from the
   parallel sorting algorithm written by Andrew Tridgell and Richard
   Brent. See ftp://samba.anu.edu.au/pub/tridge/sorting for more info
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>


/* specialise it for integer sorting */
#define INTEGER 1

typedef int BOOL;
#define False 0
#define True 1

#ifndef MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif


static int comparisons;

#if INTEGER
#define COMPARE(p1,p2) (comparisons++, *((int *)(p1)) - *((int *)(p2)))
#define ELEMENT(ptr,i) ((void *)((int *)(ptr) + (i)))
#define COPYEL(p1,p2) (*((int *)p1) = *((int *)p2))
#define COPYEL_INC(p1,p2) \
  (COPYEL(p1,p2),p1=ELEMENT(p1,1),p2=ELEMENT(p2,1))
#define COPYELS(p1,p2,n) memmove(p1,p2,(n)*EL_SIZE)
#define EL_SIZE sizeof(int)
#define PTR(p, i) ((int *)(p) + i)
#endif

#define el_vector(n) memalign(EL_SIZE,(n)*EL_SIZE)


/*******************************************************************
simple merge code - must be fast
no boundary checking is done and the number of elements taken from 
list1 is returned
********************************************************************/
static int SimpleMerge(void *dest,void *src1,void *src2,int n)
{
#if INTEGER
#define DOONE(x) (comparisons++, i1<i2?(x = i1,i1= *(++ss1)):(x = i2,i2= *(++ss2)))
{
  register int N = n;
  int *ss1=src1,*ss2=src2,*d=dest;
  int i1= *ss1,i2= *ss2;
  int d0,d1,d2,d3;
  int d4,d5,d6,d7; 

  while (N >= 8)
    {	
      DOONE(d0); DOONE(d1); DOONE(d2); DOONE(d3); 
      DOONE(d4); DOONE(d5); DOONE(d6); DOONE(d7); 
      d[0] = d0;d[1] = d1;d[2] = d2;d[3] = d3;
      d[4] = d4;d[5] = d5;d[6] = d6;d[7] = d7; 
      d+=8;N-=8;
    }

  while (N--)
    DOONE(*d++);

  /* simple_merge_sort_time += MIMD_Uptime() - tt; */
  return(ss1-(int *)src1);
#undef DOONE
}
#elif FIXED_ELEMENT
#define DOONE (COMPARE(ss1,ss2)<0? (*d++ = *ss1++) : (*d++ = *ss2++))
{
  register int N = n;
  element *ss1=src1,*ss2=src2,*d=dest;

  while (N >= 4)
    {
      DOONE; DOONE; DOONE; DOONE;
      N -= 4;
    }

  while (N--)
    DOONE;


  /* simple_merge_sort_time += MIMD_Uptime() - tt; */
  return(ss1-(element *)src1);
}
#else
{
  int N = n;
  int elsize = EL_SIZE;
  int ii;
  int (*cmp)() = PSI.comparison;

  if (elsize%sizeof(double)==0 && 
      ((int)src1)%sizeof(double)==0 && 
      ((int)src2)%sizeof(double)==0 &&
      ((int)dest)%sizeof(double)==0)
    {
      double *ss1=src1,*ss2=src2,*d=dest;
      elsize /= sizeof(double);
      switch (elsize)
	{
	case 1:
	  while (N--)
	    *d++ = (cmp(ss1,ss2)<0?*ss1++:*ss2++);
	  break;
	case 2:
	  {
	    while (N--)
	      {
		if (cmp(ss1,ss2)<0)
		  {
		    d[0] = ss1[0];d[1] = ss1[1];
		    ss1+=2;
		  }
		else
		  {
		    d[0] = ss2[0];d[1] = ss2[1];
		    ss2+=2;
		  }
		d+=2;
	      }
	  }
	  break;
	default:
	  while (N--)
	    if (cmp(ss1,ss2)<0) 
	      {ii=elsize;while (ii--) *d++ = *ss1++;}
	    else 
	      {ii=elsize;while (ii--) *d++ = *ss2++;}
	  break;
	}
      /* simple_merge_sort_time += MIMD_Uptime() - tt; */
      return(((char *)ss1 - (char *)src1)/EL_SIZE);
    }

  if (elsize==sizeof(float) && 
      ((int)src1)%sizeof(float)==0 && 
      ((int)src2)%sizeof(float)==0 &&
      ((int)dest)%sizeof(float)==0)
    {
      float *ss1=src1,*ss2=src2,*d=dest;
      elsize /= sizeof(float);
      while (N--)
	*d++ = (cmp(ss1,ss2)<0?*ss1++:*ss2++);
      /* simple_merge_sort_time += MIMD_Uptime() - tt; */
      return(((char *)ss1 - (char *)src1)/EL_SIZE);
    }

    {
      char *ss1=src1,*ss2=src2,*d=dest;
      while (N--)
	{
	  if (cmp(ss1,ss2)<0)
	    {
	      memmove(d,ss1,elsize);
	      d+=elsize;
	      ss1+=elsize;
	    }
	  else
	    {
	      memmove(d,ss2,elsize);
	      d+=elsize;
	      ss2+=elsize;
	    }
	}
      /* simple_merge_sort_time += MIMD_Uptime() - tt; */
      return(((char *)ss1 - (char *)src1)/EL_SIZE);
    }
}
#endif
}



/*******************************************************************
shift some elements - may overlap!
********************************************************************/
static void ShiftEls(void *dest,void *src,int N)
{
  memmove(PTR(dest,0),PTR(src,0),N*EL_SIZE);
}

/*******************************************************************
unbalanced merge of two lists - best when L1 or L2 is small
********************************************************************/
static void UnbalancedMerge(void *data,int L1,int L2,int gap)
{
  int *slots;
  int N = L1 + L2;
  int smallL = MIN(L1,L2);
  int bigL = MAX(L1,L2);
  void *smallP,*bigP;

  slots = memalign(EL_SIZE,sizeof(int)*smallL);
  
  if (L1 < L2) 
    {
      smallP = data;
      bigP = PTR(data,L1+gap);
    }
  else
    {
      bigP = data;
      smallP = PTR(data,L1+gap);
    }
  
  /* create the slots list - this says where each element in 
     smallList will fit into the big list */
  {
    int start=0;
    int i;
    for (i=0;i<smallL;i++)
      {
	int min=start,max=bigL,proposed;
	while (min<max)
	  {
	    proposed = (min+max)/2 + 1;
	    if (COMPARE(PTR(smallP,i),PTR(bigP,proposed-1))<0)
	      max = proposed-1;
	    else
	      min = proposed;
	  }

	slots[i] = min;
	start = min;
      }
  }
  
  /* now fit the elements into the bigger list */
  if (L1 < L2)
    {
      int slot=0;
      int num_taken=0;
      int num_done=0;
      void *TempBuf = memalign(EL_SIZE,smallL*EL_SIZE);
      COPYELS(TempBuf,smallP,smallL);

      while (slot<smallL)
	{
	  int todo = slots[slot]-num_taken;
	  ShiftEls(PTR(data,num_done),PTR(bigP,num_taken),todo);
	  num_done += todo;
	  num_taken += todo;
	  COPYEL(PTR(data,num_done),PTR(TempBuf,num_done-num_taken));
	  num_done++;
	  slot++;
	}
      if (num_done < N)
	ShiftEls(PTR(data,num_done),PTR(bigP,num_taken),N-num_done);
      free(TempBuf);
    }
  else
    {
      int slot=smallL-1;
      int num_taken=0;
      int num_done=0;
      void *TempBuf = memalign(EL_SIZE,smallL*EL_SIZE);
      COPYELS(TempBuf,smallP,smallL);

      while (slot>=0)
	{
	  int todo = bigL-slots[slot]-num_taken;
	  ShiftEls(PTR(data,N-num_done-todo),
		   PTR(bigP,bigL-num_taken-todo),
		   todo);
	  num_done += todo;
	  num_taken += todo;
	  COPYEL(PTR(data,N-num_done-1),
		 PTR(TempBuf,smallL-(num_done-num_taken)-1));
	  num_done++;
	  slot--;
	}
      ShiftEls(data,bigP,N-num_done);
      free(TempBuf);
    }


  free(slots);
}


/*******************************************************************
merge 2 lists into one with minimum memory overhead
********************************************************************/
static void insitu_merge(void *data,int L1,int L2,int gap)
{
  int N = L1 + L2;
  int B;
  int bListLen1,bListLen2,bListLen3;
  int NumTemp=3;
  int TailSize;
  BOOL TailDone;
  typedef struct
    {
      int start,length;
    } bListStruct;
  
  bListStruct *bList1,*bList2,*bList3;
  void *TailBuf;
  void *TempBuf;
  double sqrt(double );
  
  
#define COPYP(p1,p2) COPYEL(p1,p2)
#define COPYP_INC(p1,p2) COPYEL_INC(p1,p2)
#define COPYPN(p1,p2,n) COPYELS(p1,p2,n)
#define LESSTHANP(p1,p2) (COMPARE(p1,p2)<0)

  if (L2==0) return;
  
  if (L1==0 && gap>0)
    {
      ShiftEls(data,PTR(data,gap),L2);
      return;
    }

  B = MIN(N,(int)sqrt((double)N)*4);
  /* B = MIN(B,(128*1024/EL_SIZE)/4); */
  
  
#if UNBALANCED_MERGE
  if (L1 <= B || L2 <= B)
    {
      UnbalancedMerge(data,L1,L2,gap);
      return;
    }
#endif
  
  /* INITIALISATION: allocate memory */
  TempBuf = el_vector(NumTemp*B);
  
  bList1 = (bListStruct *)memalign(EL_SIZE,(L1/B + 2)*sizeof(bListStruct));
  bList2 = (bListStruct *)memalign(EL_SIZE,(L2/B + 2)*sizeof(bListStruct));
  bList3 = (bListStruct *)memalign(EL_SIZE,(L1/B + L2/B + 5)*sizeof(bListStruct));
  
  TailSize = L1 % B;
  TailBuf = el_vector(TailSize);
  TailDone = (TailSize == 0);
  

  /* STAGE 1: Create the 2 source block lists and initialise the 3rd */
  {
    int L,S,pos;
    bListStruct *list;
    
    list=bList1; L=L1 - TailSize; S=0; pos=0;
    while (L)
      {
	list[pos].start = S + pos * B;
	list[pos].length = MIN(B,L);
	L -= list[pos].length;
	pos++;
      }
    bListLen1=pos;
    
    /* copy the tail from list 1 */
    COPYPN(TailBuf,PTR(data,bListLen1*B),TailSize);
    
    list=bList2; L=L2; S=L1+gap; pos=0;
    while (L)
      {
	list[pos].start = S + pos * B;
	list[pos].length = MIN(B,L);
	L -= list[pos].length;
	pos++;
      }
    bListLen2=pos;
    
    bListLen3=0;
  }
  
  /* STAGE 2: merge the two lists into list3 */
  {
    void *pL1,*pL2,*pL3;
    int avail1,avail2;
    int space3;
    int sPos1,sPos2;
    int availPos3,dPos3;
    BOOL empty1,empty2;
    
    /* select the first blocks in list 1 and 2 */
    if (bListLen1 > 0)
      {
	pL1 = PTR(data,bList1[0].start);
	avail1 = bList1[0].length;
	sPos1 = 0;
	empty1 = False;
      }
    else
      {
	pL1 = TailBuf;
	avail1 = TailSize;
	TailDone = True;
	sPos1 = 0;
	empty1 = False;
      }
    
    pL2 = PTR(data,bList2[0].start);
    avail2 = bList2[0].length;
    sPos2 = 0;
    empty2 = False;
    
    /* initially select the TempBuf */
    bList3[0].start = -1;
    bList3[0].length = NumTemp*B;
    availPos3 = 1;
    pL3 = TempBuf;
    space3 = NumTemp*B;
    dPos3 = 0;
    
    
    while (!empty1 || !empty2)
      {
	int MinToDo = space3;
	if (!empty1) MinToDo = MIN(MinToDo,avail1);
	if (!empty2) MinToDo = MIN(MinToDo,avail2);
	
	space3 -= MinToDo;
	if (empty2 || empty1)
	  {
	    if (empty1)
	      {
		COPYPN(pL3,pL2,MinToDo);
		pL2 = PTR(pL2,MinToDo);
		avail2 -= MinToDo;
	      }
	    else
	      {
		COPYPN(pL3,pL1,MinToDo);
		pL1 = PTR(pL1,MinToDo);
		avail1 -= MinToDo;
	      }
	    pL3 = PTR(pL3,MinToDo);
	  }
	else
	  {
	    int FrompL1 = SimpleMerge(pL3,pL1,pL2,MinToDo);
	    avail1 -= FrompL1;
	    avail2 -= (MinToDo-FrompL1);
	    pL3 = PTR(pL3,MinToDo);
	    pL1 = PTR(pL1,FrompL1);
	    pL2 = PTR(pL2,MinToDo-FrompL1);
	  }
	
	/* have I emptied list 1 ? */
	if (avail1==0 && !empty1)
	  {
	    /* make the just emptied block available */
	    if (TailSize == 0 || !TailDone)
	      {
		/*	      int k = dPos3+1; */
		int k = availPos3;
		
		memmove(&bList3[k+1],&bList3[k],(availPos3-k)*sizeof(bList3[k]));
		
		bList3[k].start = bList1[sPos1].start;
		bList3[k].length = bList1[sPos1].length;
		availPos3++;
		sPos1++;
	      }
	    
	    /* are there no blocks left ? */
	    if (sPos1 == bListLen1) 
	      {
		if (!TailDone)
		  {
		    pL1 = TailBuf;
		    avail1 = TailSize;
		    TailDone = True;
		  }
		else
		  empty1 = True;
	      }
	    else
	      {
		/* start using the next block */
		pL1 = PTR(data,bList1[sPos1].start);
		avail1 = bList1[sPos1].length;
	      }
	  }
	
	/* have I emptied list 2 ? */
	if (avail2==0 && !empty2)
	  {
	    /*	  int k = dPos3+1; */
	    int k = availPos3;
	    
	    memmove(&bList3[k+1],&bList3[k],(availPos3 - k)*sizeof(bList3[k]));
	    
	    /* make the just emptied block available */	
	    bList3[k].start = bList2[sPos2].start - (TailSize + gap);
	    bList3[k].length = bList2[sPos2].length;	
	    if (sPos2 == (bListLen2-1))
	      bList3[k].length += TailSize;
	    if (bList3[k].length > B)
	      {
		bList3[availPos3+1].length = bList3[k].length - B;
		bList3[availPos3+1].start = bList3[k].start + B;
		bList3[k].length = B;
		availPos3++;
	      }		
	    
	    /* don't make odd sized blocks available */
	    if (bList3[availPos3].length == B)
	      availPos3++;
	    sPos2++;
	    
	    /* are there no blocks left ? */
	    if (sPos2 == bListLen2) 
	      empty2 = True;
	    else
	      {
		/* start using the next block */
		pL2 = PTR(data,bList2[sPos2].start);
		avail2 = bList2[sPos2].length;
	      }
	  }
	
	/* have I filled list 3 ? */
	if (space3 == 0)
	  {
	    /* start using the next block */
	    dPos3++;
	    if (dPos3 == availPos3)
	      printf("\nmerge error - dPos3==availPos3! %d\n",dPos3);
	    pL3 = PTR(data,bList3[dPos3].start);
	    space3 = bList3[dPos3].length;
	  }
      }
    
    /* we may not have filled list 3 completely */
    bList3[dPos3].length -= space3;
    
    /* move the last elements to the last block in list 2 */
    COPYPN(PTR(data,N-bList3[dPos3].length),
	   PTR(pL3,-bList3[dPos3].length),
	   bList3[dPos3].length);
    bList3[dPos3].start = N - bList3[dPos3].length;
    
    bListLen3 = dPos3;
  }

  /* STAGE 3: re-arrange to give final result */
  {
    typedef struct
      {
	int where_hiding;
	int what_now;
	BOOL occupied;
      } block_struct;
    block_struct *blocks;
    int num_blocks = N / B;
    int i;
    BOOL all_done = False;
    
    blocks = (block_struct *)memalign(EL_SIZE,num_blocks * sizeof(block_struct));
    for (i=0;i<num_blocks;i++)
      blocks[i].occupied = False;
    /* the temp blocks first */
    for (i=0;i<MIN(num_blocks,NumTemp);i++)
      blocks[i].where_hiding = -(i+1);
    for (i=1;i<bListLen3;i++)
      {
	blocks[i-1+NumTemp].where_hiding = bList3[i].start/B;
	blocks[blocks[i-1+NumTemp].where_hiding].what_now = i-1+NumTemp;
	blocks[blocks[i-1+NumTemp].where_hiding].occupied = True;
      }
    
    while (!all_done)
      {
	BOOL no_holes = False;
	all_done = True;
	while (!no_holes)
	  {
	    no_holes = True;
	    for (i=0;i<num_blocks;i++)
	      {
		while (!blocks[i].occupied)
		  {
		    int from = blocks[i].where_hiding;
		    no_holes = False;
		    if (from >= 0)
		      {
			COPYPN(PTR(data,i*B),PTR(data,from*B),B);
			blocks[from].occupied = False;
		      }
		    else
		      COPYPN(PTR(data,i*B),PTR(TempBuf,-(from+1)*B),B);
		  
		    blocks[i].occupied = True;
		    blocks[i].what_now = i;
		    blocks[i].where_hiding = i;		  
		    if (from >= 0)
		      i = from;
		  }
	      }
	  }
	for (i=0;i<num_blocks && all_done;i++)
	  if (blocks[i].what_now != i)
	    {
	      all_done = False;
	      COPYPN(TempBuf,PTR(data,i*B),B);
	      blocks[i].occupied = False;
	      blocks[blocks[i].what_now].where_hiding = -1;	    
	    }
      }
    free(blocks);
  }
  free(TempBuf);
  free(bList1);
  free(bList2);
  free(bList3);
  if (TailBuf) free(TailBuf);
}


/* merge two sorted lists of elements into one list. The two lists are
   assumed to be contiguous with the first starting at data[0] and the
   second starting at data[n1]. The two lists are of length n1 and n2
   respectively */
int imerge(int *data, int n1, int n2)
{
	comparisons=0;
	insitu_merge(data, n1, n2, 0);
	return comparisons;
}
