/* a simple test harness to test integer list merging algorithms 

   currently setup to test ShuffleMerge, imerge and LangstonMerge 

   Andrew.Tridgell@anu.edu.au March 1998
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <sys/resource.h>


/* some useful accurate timing routines  */
struct rusage tp1,tp2;

static void start_timer()
{
	getrusage(RUSAGE_SELF,&tp1);
}


/* returns the time in seconds since start_timer was called */
static double end_timer()
{
	getrusage(RUSAGE_SELF,&tp2);
	return((tp2.ru_utime.tv_sec - tp1.ru_utime.tv_sec) + 
	       (tp2.ru_utime.tv_usec - tp1.ru_utime.tv_usec)*1.0e-6);
}


/* call this when out of memory */
static void oom(void)
{
	printf("out of memory\n");
	exit(1);
}

/* minimum time for any size list - ensures reasonably accurate timings */
#define MIN_TIME 2

/* create a random list of integers of length n */
static int *rand_list(int n)
{
	int *ret = (int *)malloc(n * sizeof(int));
	int i;

	if (!ret) oom();
	for (i=0;i<n;i++)
		ret[i] = random();
	return ret;
}


/* used by qsort */
static int int_comparison(int *i1, int *i2)
{
	return *i1 - *i2;
}

/* a simple sort routine using qsort - used to establish the lists */
static void sort_list(int *data, int n)
{
	qsort(data, n, sizeof(int), int_comparison);
}


static void usage(void)
{
	printf("testmerge <start_n> <end_n> <multiplier>\n");
}


int imerge(int *data, int n1, int n2);
int ShuffleMerge(int *data, int n1, int n2);
int LangstonMerge(int *data, int n1, int n2);
int dmerge(int *data, int n1, int n2);

int main(int argc, char *argv[])
{
	int start_n, end_n, n, loops;
	int *data, *data2, *data3, *data4, *data5, *data6;
	double t1, t2, t3, t4, t5, multiplier;
	int c1, c2, c3, c4, c5;

	if (argc < 4) {
		usage();
		exit(0);
	}

	start_n = atoi(argv[1]);
	end_n = atoi(argv[2]);
	multiplier = atof(argv[3]);

	srandom(time(NULL) ^ getpid());

	printf("  n     imerge  smerge  lmerge  dmerge  smerge2\n");
	
	
	/* we loop increasing the list size by multiplier each time. 
	   for each size we create a random list and sort the 2 halves. 
	   we then test the different merge algorithms a number of times,
	   ensuring that each algorithm runs for at least MIN_TIME seconds
	   to reduce errors */
	for (n=start_n; n <= end_n; n *= multiplier) {
		/* create a random list of length n and sort the 2 halves */
		data = rand_list(n);
		sort_list(data, n/2);
		sort_list(data+n/2, n-n/2);

		data2 = malloc(sizeof(int)*n);
		data3 = malloc(sizeof(int)*n);
		data4 = malloc(sizeof(int)*n);
		data5 = malloc(sizeof(int)*n);
		data6 = malloc(sizeof(int)*n);
		if (!data2 || !data3 || !data4 || !data5 || !data6) oom();
		
		/* time the imerge */
		start_timer();
		for (t1=0, c1=0, loops=0; t1 < MIN_TIME; ) {
			memcpy(data2, data, sizeof(int)*n);
			c1 += imerge(data2, n/2, n-n/2);
			t1 = end_timer();
			loops++;
		}
		t1 /= loops;
		c1 /= loops;

		/* time the shuffle merge */
		start_timer();
		for (t2=0, c2=0, loops=0; t2 < MIN_TIME; ) {
			memcpy(data3, data, sizeof(int)*n);
			c2 += ShuffleMerge(data3, n/2, n-n/2);
			t2 = end_timer();
			loops++;
		}
		t2 /= loops;
		c2 /= loops;

		/* time the Langston merge */
		start_timer();
		for (t3=0, c3=0, loops=0; t3 < MIN_TIME; ) {
			memcpy(data4, data, sizeof(int)*n);
			c3 += LangstonMerge(data4, n/2, n-n/2);
			t3 = end_timer();
			loops++;
		}
		t3 /= loops;
		c3 /= loops;

		/* time the dumb merge */
		start_timer();
		for (t4=0, c4=0, loops=0; t4 < MIN_TIME; ) {
			memcpy(data5, data, sizeof(int)*n);
			c4 += dmerge(data5, n/2, n-n/2);
			t4 = end_timer();
			loops++;
		}
		t4 /= loops;
		c4 /= loops;

		/* time the ShuffleMerge varient */
		start_timer();
		for (t5=0, c5=0, loops=0; t5 < MIN_TIME; ) {
			memcpy(data6, data, sizeof(int)*n);
			c5 += ShuffleMerge2(data6, n/2, n-n/2);
			t5 = end_timer();
			loops++;
		}
		t5 /= loops;
		c5 /= loops;

		/* make sure they all got the same result! */
		if (memcmp(data2, data3, sizeof(int)*n) != 0) {
			printf("data2 and data3 don't match!\n");
			exit(1);
		}

		if (memcmp(data2, data4, sizeof(int)*n) != 0) {
			printf("data2 and data4 don't match!\n");
			exit(1);
		}

		if (memcmp(data2, data5, sizeof(int)*n) != 0) {
			printf("data2 and data5 don't match!\n");
			exit(1);
		}

		if (memcmp(data2, data6, sizeof(int)*n) != 0) {
			printf("data2 and data6 don't match!\n");
			exit(1);
		}

#if 0
		printf("%7d\t%.3g\t%.3g\t%.3g\t%.3g\t%.3g\n", 
		       n, t1, t2, t3, t4, t5);
#else
		printf("%7d\t%.3g\t%.3g\t%.3g\t%.3g\t%.3g\n", n, 
		       c1/(double)n,
		       c2/(double)n,
		       c3/(double)n,
		       c4/(double)n,
		       c5/(double)n);
#endif

		free(data);
		free(data2);
		free(data3);
		free(data4);
		free(data5);
	}
	
	return 0;
}
